package jp.topse.kaizu.sequence2csp.convert;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

public abstract class SequenceTree {
	private enum LastType {
		SEND, RECEIVE
	}

	public static SequenceTree create() {
		return new SequenceTreeMessage(null);
	}

	public abstract SequenceTree addMessage(String messageString,
			boolean receive);

	public abstract void addState(String compoString, String stateString);

	public abstract String toString(int indent, LastType type);

	public String toString() {
		return toString(1, LastType.SEND);
	}

	protected void printIndent(StringBuilder sb, int indent) {
		for (int i = 0; i < indent; i++) {
			sb.append("\t");
		}
	}

	private static class SequenceTreeMessage extends SequenceTree {
		private String messageString;

		private Map<String, SequenceTree> sendChildren;

		private Map<String, SequenceTree> receiveChildren;

		public SequenceTreeMessage(String messageString) {
			this.messageString = messageString;
			this.sendChildren = new HashMap<String, SequenceTree>();
			this.receiveChildren = new HashMap<String, SequenceTree>();
		}

		@Override
		public SequenceTree addMessage(String message, boolean receive) {
			Map<String, SequenceTree> children = (receive ? receiveChildren
					: sendChildren);
			if (!children.containsKey(message)) {
				children.put(message, new SequenceTreeMessage(message));
			}
			return children.get(message);
		}

		private boolean isRoot() {
			return (messageString == null);
		}

		@Override
		public void addState(String compoString, String state) {
			if (!isRoot() && !receiveChildren.containsKey(state)) {
				receiveChildren.put(state, new SequenceTreeState(compoString,
						state));
			}
		}

		@Override
		public String toString(int indent, LastType type) {
			StringBuilder sb = new StringBuilder();
			boolean selection = (sendChildren.size() + receiveChildren.size() > 1);
			boolean timeout = (sendChildren.size() > 0 && receiveChildren
					.size() > 0);
			if (messageString != null) {
				sb.append(messageString);
				sb.append(" ->");
				sb.append(Util.LINE_SEPARATOR);
				if (selection) {
					indent++;
				}
				printIndent(sb, indent);
			}

			appendIf(sb, timeout, "(");
			if (type == LastType.SEND) {
				appendSendChildren(sb, indent, sendChildren);
				appendChoice(sb, indent, timeout, "[>");
				appendReceiveChildren(sb, indent, receiveChildren);
			} else {
				appendReceiveChildren(sb, indent, receiveChildren);
				appendChoice(sb, indent, timeout, "[>");
				appendSendChildren(sb, indent, sendChildren);
			}
			appendIf(sb, timeout, ")");

			if (messageString != null && selection) {
				indent--;
			}
			return sb.toString();
		}

		private void appendIf(StringBuilder sb, boolean cond, String data) {
			if (cond) {
				sb.append(data);
			}
		}

		private void appendSendChildren(StringBuilder sb, int indent,
				Map<?, SequenceTree> children) {
			appendChildren(sb, indent, LastType.SEND, children.values(), "|~|");
		}

		private void appendReceiveChildren(StringBuilder sb, int indent,
				Map<?, SequenceTree> children) {
			appendChildren(sb, indent, LastType.RECEIVE, children.values(),
					"[]");
		}

		private void appendChildren(StringBuilder sb, int indent,
				LastType type, Collection<SequenceTree> children, String choice) {
			appendIf(sb, children.size() > 1, "(");
			boolean notfirst = false;
			for (SequenceTree child : children) {
				appendChoice(sb, indent, notfirst, choice);
				sb.append(child.toString(indent, type));
				notfirst = true;
			}
			appendIf(sb, children.size() > 1, ")");
		}

		private void appendChoice(StringBuilder sb, int indent, boolean cond,
				String choice) {
			if (cond) {
				sb.append(Util.LINE_SEPARATOR);
				printIndent(sb, indent);
				sb.append(choice);
				sb.append(Util.LINE_SEPARATOR);
				printIndent(sb, indent);
			}
		}
	}

	private static class SequenceTreeState extends SequenceTree {
		public String compoString;

		public String stateString;

		public SequenceTreeState(String compoString, String stateString) {
			this.compoString = compoString;
			this.stateString = stateString;
		}

		@Override
		public SequenceTree addMessage(String message, boolean receive) {
			throw new RuntimeException("Internal Error");
		}

		@Override
		public void addState(String compoString, String state) {
			throw new RuntimeException("Internal Error");
		}

		@Override
		public String toString(int indent, LastType type) {
			return Converter.COMPO_KEYWORD + "(" + compoString + "."
					+ stateString + ")";
		}
	}
}
